#!/usr/bin/env python
#title           :app.py
#description     :This file is used to start the PyVuka commandline program
#                 through a python import with clear syntax
#usage           :import PyVuka.app as app
#                 app.start()
#python_version  :3.7
#==============================================================================
from . import pyvuka


def start():
    pyvuka.start()


